//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  SocketCAN high level access by Zeljko Avramovic (c) 2021                            //
//                                                                                      //
//  Dual license:                                                                       //
//  1. FPC modified LGPL (chosen for compatibility with FreePascal and Lazarus)         //
//  2. BSD3              (chosen for compatibility with everything else)                //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

unit can.hl;

{$mode objfpc}
{$modeswitch typehelpers}
{$modeswitch advancedrecords}
{$packrecords c}
{$inline on}

interface

uses
  baseunix, sockets, strings, strutils, can, can.bcm;

// +=====+=====+=====+============================================+
// | 31  | 30  | 29  |                   28..0                    |
// |-----+-----+-----+--------------------------------------------|
// | EFF | RTR | ERR |                   CAN ID                   |   <= Extended CAN ID
// +=====+=====+=====+============================================+
// | 31  | 30  | 29  | 28..26   |           25..8          | 7..0 |
// |-----+-----+-----+----------+--------------------------+------|
// | EFF | RTR | ERR | PRIORITY |            PGN           |  SA  |   <= J1939 with Extended CAN ID details
// +=====+=====+=====+==========+==========================+======+
// | 31  | 30  | 29  | 28..26   | 25 | 24 | 23..16 | 15..8 | 7..0 |
// |-----+-----+-----+----------+----+----+--------+-------+------|
// | EFF | RTR | ERR | PRIORITY | R  | DP |   PF   |  PS   |  SA  |   <= J1939 with PGN details
// +=====+=====+=====+==========+====+====+========+=======+======+

const
  S_IRUSR =  1 shl 3;

  CAN_MAX_CAN_ID_NUMBER = (1 shl 29)-1;
  J1939_MAX_PGN_NUMBER  = (1 shl 18)-1;

type
  TCanSimpleID = 0..CAN_MAX_CAN_ID_NUMBER; // to access whole value of 29 bit CAN ID

  TCan32Bits = bitpacked array[0..31] of boolean;

  // Controller Area Network Error Message Frame Mask structure
  // bit 0-28	: error class mask (see include/uapi/linux/can/error.h)
  // bit 29-31	: set to zero
  // error class (mask) in can_id
  TCanErrMask = bitpacked record // high level for can_err_mask_t
    private
      function  GetSFF: boolean;
      procedure SetSFF(const ANewMaskSFF: boolean);
      function  GetEFF: boolean;
      procedure SetEFF(const ANewMaskEFF: boolean);
    public
      property  SFF: boolean read GetSFF write SetSFF; // CAN_SFF_MASK = $000007FF, helper to set standard frame format (SFF)
      property  EFF: boolean read GetEFF write SetEFF; // CAN_EFF_MASK = $1FFFFFFF, helper to set extended frame format (EFF)
      property  ERR: boolean read GetEFF write SetEFF; // CAN_ERR_MASK = $1FFFFFFF, helper to omit EFF, RTR and ERR flags

    case byte of
      1: ( Error: record
             TxTimeout:       boolean; // CAN_ERR_TX_TIMEOUT = $00000001; // TX timeout (by netdevice driver)
             LostArbitration: boolean; // CAN_ERR_LOSTARB    = $00000002; // lost arbitration    / data[0]
             Controller:      boolean; // CAN_ERR_CRTL       = $00000004; // controller problems / data[1]
             Protocol:        boolean; // CAN_ERR_PROT       = $00000008; // protocol violations / data[2..3]
             Transceiver:     boolean; // CAN_ERR_TRX        = $00000010; // transceiver status  / data[4]
             NoAck:           boolean; // CAN_ERR_ACK        = $00000020; // received no ACK on transmission
             BusOff:          boolean; // CAN_ERR_BUSOFF     = $00000040; // bus off
             BusError:        boolean; // CAN_ERR_BUSERROR   = $00000080; // bus error (may flood!)
             Restarted:       boolean; // CAN_ERR_RESTARTED  = $00000100; // controller restarted
             __Dummy:         0..(1 shl 20)-1;                            // 20 bits not defined yet
             Zeros:           0..7;                                       // set to zero
           end; );
      2: ( Flag: record
             __Dummy:       TCanSimpleID;  // trick to skip until we reach wanted bits
             ERR:           boolean;       // bit 29   : error message frame flag (0 = data frame, 1 = error message)
             RTR:           boolean;       // bit 30   : remote transmission request flag (1 = rtr frame)
             EFF:           boolean;       // bit 31   : extended frame format flag (0 = standard 11 bit, 1 = extended 29 bit)
           end; );
      3: ( Bit:           TCan32Bits );    // bit access helper
      4: ( Value:         TCanSimpleID );  // access helper for first 29 bits
      5: ( Full:          cuint32 );       // full access helper
  end;

  // Controller Area Network Identifier structure
  TCanID = bitpacked record case byte of // high level for canid_t
    1: ( Value:         TCanSimpleID;    // bit 0-28 : CAN identifier (11/29 bit)
         ERR:           boolean;         // bit 29   : error message frame flag (0 = data frame, 1 = error message)
         RTR:           boolean;         // bit 30   : remote transmission request flag (1 = rtr frame)
         EFF:           boolean; );      // bit 31   : extended frame format flag (0 = standard 11 bit, 1 = extended 29 bit)
    2: ( Mask:          TCanErrMask );   // overlay for error and flag bits
    3: ( __Dummy:       TCanSimpleID;    // trick to skip bits until we reach InverseFilter bit
         InverseFilter: boolean; );      // CAN_INV_FILTER bit 29 ($20000000) only used for filtering frames in TCanFrame.ID;
    4: ( Bit:           TCan32Bits );    // bit access helper
    5: ( Full:          cuint32 );       // full access helper
  end;

  // CAN ID based filter in can_register()
  TCanFilter = record  // high level for can_filter
    ID:   TCanID;      // relevant bits of CAN ID which are not masked out
    Mask: TCanErrMask; // CAN mask
                       // A filter matches, when <received_can_id> and Mask = ID and Mask
                       // The filter can be inverted (CAN_INV_FILTER bit set in ID) or it can
                       // filter for error message frames (CAN_ERR_FLAG bit set in mask).
  end;

  TCanDataErrors = bitpacked record  // detailed error bits received in CAN frame data[]
    //                               // for details look at can.error.pas
    LostArbitration: record          // arbitration lost in bit ...                  / data[0]
      BitNumber:            cuint8;  // unspecified if CAN_ERR_LOSTARB_UNSPEC = $00
    end;                             // else bit number in bitstream
    //
    Controller: record               // error status of CAN-controller               / data[1]
                                     // CAN_ERR_CRTL_UNSPEC      = $00, unspecified
      OverflowRX:           boolean; // CAN_ERR_CRTL_RX_OVERFLOW = $01, RX buffer overflow
      OverflowTX:           boolean; // CAN_ERR_CRTL_TX_OVERFLOW = $02, TX buffer overflow
      WarningRX:            boolean; // CAN_ERR_CRTL_RX_WARNING  = $04, reached warning level for RX errors
      WarningTX:            boolean; // CAN_ERR_CRTL_TX_WARNING  = $08, reached warning level for TX errors
      PassiveRX:            boolean; // CAN_ERR_CRTL_RX_PASSIVE  = $10, reached error passive status RX
      PassiveTX:            boolean; // CAN_ERR_CRTL_TX_PASSIVE  = $20, reached error passive status TX
                                           // at least one error counter exceeds the protocol-defined level of 127
      Active:               boolean; // CAN_ERR_CRTL_ACTIVE      = $40, recovered to error active state
    end;
    //
    Protocol: record                 // error in CAN protocol
      Category: record               // error in CAN protocol (type)                 / data[2]
                                     // CAN_ERR_PROT_UNSPEC      = $00, unspecified
        SingleBit:          boolean; // CAN_ERR_PROT_BIT         = $01, single bit error
        FrameFormat:        boolean; // CAN_ERR_PROT_FORM        = $02, frame format error
        BitStuffing:        boolean; // CAN_ERR_PROT_STUFF       = $04, bit stuffing error
        Bit0:               boolean; // CAN_ERR_PROT_BIT0        = $08, unable to send dominant bit
        Bit1:               boolean; // CAN_ERR_PROT_BIT1        = $10, unable to send recessive bit
        BusOverload:        boolean; // CAN_ERR_PROT_OVERLOAD    = $20, bus overload
        ActiveAnnouncement: boolean; // CAN_ERR_PROT_ACTIVE      = $40, active error announcement
        TX:                 boolean; // CAN_ERR_PROT_TX          = $80, error occurred on transmission
      end;
      //
      Location:             cuint8;  // error in CAN protocol (location)             / data[3]
                                     // CAN_ERR_PROT_LOC_UNSPEC        = $00, unspecified
                                     // CAN_ERR_PROT_LOC_SOF           = $03, start of frame
                                     // CAN_ERR_PROT_LOC_ID28_21       = $02, ID bits 28 - 21 (SFF: 10 - 3)
                                     // CAN_ERR_PROT_LOC_ID20_18       = $06, ID bits 20 - 18 (SFF: 2 - 0 )
                                     // CAN_ERR_PROT_LOC_SRTR          = $04, substitute RTR (SFF: RTR)
                                     // CAN_ERR_PROT_LOC_IDE           = $05, identifier extension
                                     // CAN_ERR_PROT_LOC_ID17_13       = $07, ID bits 17-13
                                     // CAN_ERR_PROT_LOC_ID12_05       = $0F, ID bits 12-5
                                     // CAN_ERR_PROT_LOC_ID04_00       = $0E, ID bits 4-0
                                     // CAN_ERR_PROT_LOC_RTR           = $0C, RTR
                                     // CAN_ERR_PROT_LOC_RES1          = $0D, reserved bit 1
                                     // CAN_ERR_PROT_LOC_RES0          = $09, reserved bit 0
                                     // CAN_ERR_PROT_LOC_DLC           = $0B, data length code
                                     // CAN_ERR_PROT_LOC_DATA          = $0A, data section
                                     // CAN_ERR_PROT_LOC_CRC_SEQ       = $08, CRC sequence
                                     // CAN_ERR_PROT_LOC_CRC_DEL       = $18, CRC delimiter
                                     // CAN_ERR_PROT_LOC_ACK           = $19, ACK slot
                                     // CAN_ERR_PROT_LOC_ACK_DEL       = $1B, ACK delimiter
                                     // CAN_ERR_PROT_LOC_EOF           = $1A, end of frame
                                     // CAN_ERR_PROT_LOC_INTERM        = $12, intermission
    end;
    Transceiver:            cuint8;  // error status of CAN-transceiver              / data[4]
                                     // CAN_ERR_TRX_UNSPEC             = $00, 0000 0000
                                     // CAN_ERR_TRX_CANH_NO_WIRE       = $04, 0000 0100
                                     // CAN_ERR_TRX_CANH_SHORT_TO_BAT  = $05, 0000 0101
                                     // CAN_ERR_TRX_CANH_SHORT_TO_VCC  = $06, 0000 0110
                                     // CAN_ERR_TRX_CANH_SHORT_TO_GND  = $07, 0000 0111
                                     // CAN_ERR_TRX_CANL_NO_WIRE       = $40, 0100 0000
                                     // CAN_ERR_TRX_CANL_SHORT_TO_BAT  = $50, 0101 0000
                                     // CAN_ERR_TRX_CANL_SHORT_TO_VCC  = $60, 0110 0000
                                     // CAN_ERR_TRX_CANL_SHORT_TO_GND  = $70, 0111 0000
                                     // CAN_ERR_TRX_CANL_SHORT_TO_CANH = $80, 1000 0000

    ControllerSpecific: array[0..2] of cuint8; // controller specific additional info / data[5..7]
  end; // TCanDataErrors

  {$PACKRECORDS 8}
    // Basic CAN frame structure
    TCanFrame = record // high level for can_frame (originally not bitpacked)
      procedure      Clear;
      class operator Initialize(var ACanFrame: TCanFrame);
      public           // trick to make compiler accept fields after methods
      ID:     TCanID;  // 32 bit CAN_ID + EFF/RTR/ERR flags
      DLC:    cuint8;  // frame payload length in byte (0 .. CAN_MAX_DLEN)
      __pad:  cuint8;  // padding
      __res0: cuint8;  // reserved / padding
      __res1: cuint8;  // reserved / padding
      case byte of
        1: (Data:      array[0..CAN_MAX_DLEN-1]         of cuint8);  // originally defined in can.h
        2: (DataText:  array[0..CAN_MAX_DLEN-1]         of char);    // added helper for string assignment
        3: (DataByte:  array[0..CAN_MAX_DLEN-1]         of byte);    // added helper for byte assignment (same as cuint8)
        4: (DataWord:  array[0..(CAN_MAX_DLEN div 2)-1] of word);    // added helper for word assignment
        5: (DataDWord: array[0..(CAN_MAX_DLEN div 4)-1] of dword);   // added helper for double word assignment
        6: (DataQWord: array[0..(CAN_MAX_DLEN div 8)-1] of qword);   // added helper for quad word assignment
        7: (DataError: TCanDataErrors);                              // added helper for errors
    end;
  {$PACKRECORDS C}

  // defined bits for TCanFDFrame.Flags (original explanation from can.h)
  //
  // The use of struct TCanFDFrame implies the Extended Data Length (EDL) bit to
  // be set in the CAN frame bitstream on the wire. The EDL bit switch turns
  // the CAN controllers bitstream processor into the CAN FD mode which creates
  // two new options within the CAN FD frame specification:
  //
  // Bit Rate Switch - to indicate a second bitrate is/was used for the payload
  // Error State Indicator - represents the error state of the transmitting node
  //
  // As the ESI bit is internally generated by the transmitting CAN
  // controller only the BRS bit is relevant for real CAN controllers when
  // building a CAN FD frame for transmission. Setting the ESI bit can make
  // sense for virtual CAN interfaces to test applications with echoed frames.
  //
  TCanFDFlags = bitpacked record // added helper for TCanFDFrame.Flags
    BRS: boolean; // CANFD_BRS = 0x01, bit rate switch (second bitrate for payload data)
    ESI: boolean; // CANFD_ESI = 0x02, error state indicator of the transmitting node
  end;

  {$PACKRECORDS 8}
    // CAN flexible data rate frame structure
    TCanFDFrame = record   // high level for canfd_frame
      procedure Clear;
      class operator Initialize(var ACanFDFrame: TCanFDFrame);
      public // trick to make compiler accept fields after methods

      ID:     TCanID;      // 32 bit CAN_ID + EFF/RTR/ERR flags
      Length: cuint8;      // frame payload length in byte (0 .. CANFD_MAX_DLEN)
      Flags:  TCanFDFlags; // additional flags for CAN FD
      __res0: cuint8;      // reserved / padding
      __res1: cuint8;      // reserved / padding
      case byte of
        1: (Data:      array[0..CANFD_MAX_DLEN-1]         of cuint8);  // originally defined in can.h
        2: (DataText:  array[0..CANFD_MAX_DLEN-1]         of char);    // added helper for string assignment
        3: (DataByte:  array[0..CANFD_MAX_DLEN-1]         of byte);    // added helper for byte assignment (same as cuint8)
        4: (DataWord:  array[0..(CANFD_MAX_DLEN div 2)-1] of word);    // added helper for word assignment
        5: (DataDword: array[0..(CANFD_MAX_DLEN div 4)-1] of dword);   // added helper for double word assignment
        6: (DataQword: array[0..(CANFD_MAX_DLEN div 8)-1] of qword);   // added helper for quad word assignment
        7: (DataError: TCanDataErrors);                                // added helper for errors
    end;
  {$PACKRECORDS C}

  // SAE J1939 message format: https://www.kernel.org/doc/html/latest/networking/j1939.html

  TJ1939SimplePGN = 0..J1939_MAX_PGN_NUMBER; // to access whole value of 18 bit J1939 PGN

  TJ1939PS = record case byte of // J1939 PS (PDU specific), as used in J1939 PGN
    1: (Value: cuint8);   // full access helper
    2: (DA:    cuint8);   // DA (Destination Address), PDU1 Format (specific) (peer to peer), used when PF < 240 (0..$EF)
    3: (GE:    cuint8);   // GE (Group Extenstion),    PDU2 Format (global) (broadcast), used when PF >= 240 ($F0..$FF)
  end;

  TJ1939PGN = bitpacked record case byte of // J1939 PGN (Parameter Group Number), as used in J1939 29 bit CAN ID
    1: (Value: TJ1939SimplePGN; );          // full access helper
    2: (PS: TJ1939PS; // 8 bit: PS in PDU2 , else 0 (PDU specific)
        PF: cuint8;   // 8 bit: PF (PDU format). PDU1 (0..239) indicates a destination address in PS, PDU2 (250-255) indicates extension to PDU Format (PF).
        DP: boolean;  // 1 bit: DP (Data Page), page selector for PDU (Protocol Data Unit) Format (PF) field. Currently at 0, pointing to Page 0. Page 1 is for future purposes)
        R:  boolean); // 1 bit: R  (Reserved), should always be set to 0 when transmitting messages
  end;

  TJ1939ID = bitpacked record case byte of // J1939 mapping of 29 bit CAN ID
    1: ( SA:   cuint8;                     // SA (Source Address)
         case byte of                      // use trick to overcome bitpacked record byte alignment limitation
           1: ( PGN:      TJ1939PGN; );    // PGN's 18 bits expand to 24 bits when used in record
           2: ( __Dummy:  TJ1939SimplePGN; // so we need __Dummy just to skip PGN's 18 bits
                Priority: 0..7; ); );      // and then put these Priority 3 bits immediately after PGN's 18 bits
    2: ( ID:    TCanID; );                 // CAN ID overlay (with Bit[] and Value helpers)
  end;

  TJ1939Address = cuint8; // 1 byte address

  TJ1939Name = bitpacked record case byte of
    1: (IdentityNumber:          0..(1 shl 21)-1; // bit 00..20
        ManufacturerCode:        0..(1 shl 11)-1; // bit 21..31
        ECUInstance:             0..(1 shl  3)-1; // bit 32..34
        aFunctionInstance:       0..(1 shl  5)-1; // bit 35..39
        aFunction:               0..(1 shl  8)-1; // bit 40..47
        Reserved:                boolean;         // bit 48
        VehicleSystem:           0..(1 shl  7)-1; // bit 49-55
        VehicleSystemInstance:   0..(1 shl  4)-1; // bit 56-59
        IndustryGroup:           0..(1 shl  3)-1; // bit 60-62
        ArbitraryAddressCapable: boolean;);       // bit 63
    2: (Value: cuint64);
    3: (Text:  array[0..7] of char);              // added helper for string assignment
    4: (Byte:  array[0..7] of byte);              // added helper for byte assignment (same as cuint8)
    5: (Word:  array[0..3] of word);              // added helper for word assignment
    6: (DWord: array[0..1] of dword);             // added helper for double word assignment
    7: (QWord: array[0..0] of qword);             // added helper for quad word assignment
  end;

  // sockaddr structure for CAN sockets
  TCanSockAddress = record            // high level for sockaddr_can
    CanFamily:  __kernel_sa_family_t; // address family number AF_CAN
    CanIfIndex: cint;                 // CAN network interface index (Original type was integer. Is cint (which equals cint32) a proper translation?)
    CanAddress: record case byte of   // protocol specific address information
      1: (TP: record
            RxID:    TCanID;
            TxID:    TCanID;
          end;);
      2: (J1939: record
            Name:    TJ1939Name;      // 8 byte name when using dynamic addressing
            PGN:     TJ1939PGN;       // Parameter Group Number
            __Dummy: byte;            // PGN in original sockaddr_can is cuint32, and TJ1939PGN is bitpacked 18 bits which expand to 24 bits after record byte alignment, so we add 1 dummy byte to compensate
            Address: TJ1939Address;   // 1 byte address
          end;);
      // reserved for future CAN protocols address information
    end;
    procedure      Clear;
    class operator Initialize(var ACanSockAddress: TCanSockAddress);
  end;

  TCanInAddress = packed record    // high level for in_addr from socketsh.inc
    case byte of
       1: (Address: cuint32);      // inaddr_t = cuint32
       2: (Bytes:   packed array[1..4] of byte);
  end;
  PCanInAddress = ^TCanInAddress;

  TCanSockAddr = packed record     // high level for sockaddr from socketsh.inc
    {$ifdef SOCK_HAS_SINLEN}
      SaLength:       cuchar;
    {$endif}
    case integer of
      0: (SaFamily:   sa_family_t; // if SaLength is defined, sa_family_t is smaller
          SaData:     packed array[0..13] of cuint8);
      1: (SinFamily:  sa_family_t;
          SinPort:    cushort;
          SinAddress: TCanInAddress;
          SinZero:    packed array[0..7] of cuint8);
  end;
  PCanSockAddr = ^TCanSockAddr;    // do not rename TCanSockAddr to TCanSockAddress because it exists already

  TCanIfMap = record // high level for ifmap from can.if.inc
    MemStart:    culong;
    MemEnd:      culong;
    BaseAddress: cushort;
    IRQ:         cuint8;
    DMA:         cuint8;
    Port:        cuint8;
    (* 3 bytes spare *)
  end;
  PCanIfMap = ^TCanIfMap;

  TCanIfSettings = record // high level for if_settings from can.if.inc
  	PType: cuint; (* Type of physical device or protocol *)
  	Size:  cuint; (* Size of the data allocated by the caller *)
    IfsIfsu: record
  	  case byte of
  		  (* {atm/eth/dsl}_settings anyone ? *)
  		  1: ( RawHDLC:     Praw_hdlc_proto );
  		  2: ( Cisco:       Pcisco_proto );
  		  3: ( Fr:          Pfr_proto );
  		  4: ( FrPVC:       Pfr_proto_pvc );
  		  5: ( FrPVCInfo:   Pfr_proto_pvc_info );
  		  6: ( x25:         Px25_hdlc_proto);

  		  (* interface settings *)
  		  7: ( Sync:        Psync_serial_settings );
  		  8: ( Te1:         Pte1_settings );
    end;
  end;

  TCanIFreq = record // high level for ifreq from can.if.inc
    private
      function  GetName: string;
      procedure SetName(const AName: string);
    public
      property Name: string read GetName write SetName; // helper to allow pascal strings instead of null terminated (to use CanInterface.Name := 'vcan0' instead of CanInterface.CName := 'vcan0' + #0)

    case byte of
      1: (  // allow indirect data usage through ifr_ifrn and ifr_ifru structures
            IFRN: record
              // original field name was Name but renamed to CName because of the helper property Name
              CName: array[0..(IFNAMSIZ)-1] of char;
            end;
            IFRU: record
              case byte of
                 1: ( Address:          TCanSockAddr );
                 2: ( DSTAddress:       TCanSockAddr );
                 3: ( BroadcastAddress: TCanSockAddr );
                 4: ( NetMask:          TCanSockAddr );
                 5: ( HardwareAddress:  TCanSockAddr );
                 6: ( Flags:            cshort );
                 7: ( IValue:           cint );
                 8: ( MTU:              cint );
                 9: ( Map:              TCanIfMap );
                10: ( Slave:            array[0..(IFNAMSIZ)-1] of char );
                11: ( NewName:          array[0..(IFNAMSIZ)-1] of char );
                12: ( Data:             __caddr_t );
                13: ( Settings:         TCanIfSettings );
            end;
         );
      2: (  // allow direct data usage without ifr_ifrn and ifr_ifru structures (slightly different field naming)
            // original field name was Name but renamed to CName  because of the helper property Name
            CName: array[0..(IFNAMSIZ)-1] of char;                        (* interface name        *)
            case byte of
              1 : ( Address:            TCanSockAddr );                   (* address               *)
              2 : ( DSTAddress:         TCanSockAddr );                   (* other end of p-p lnk  *)
              3 : ( BroadcastAddress:   TCanSockAddr );                   (* broadcast address     *)
              4 : ( NetMask:            TCanSockAddr );                   (* interface net mask    *)
              5 : ( HardwareAddress:    TCanSockAddr );                   (* MAC address           *)
              6 : ( Flags:              cshort );                         (* flags                 *)
              7 : ( Metric:             cint );                           (* metric                *)
              8 : ( MTU:                cint );                           (* mtu                   *)
              9 : ( Map:                TCanIfMap );                      (* device map            *)
              10: ( Slave:              array[0..(IFNAMSIZ)-1] of char ); (* slave device          *)
              11: ( NewName:            array[0..(IFNAMSIZ)-1] of char ); (* New name              *)
              12: ( Data:               __caddr_t );                      (* for use by interface  *)
              13: ( Bandwidth:          cint );                           (* link bandwidth        *)
              14: ( IfIndex:            cint );                           (* interface index       *)
              15: ( QLength:            cint );                           (* Queue length          *)
              16: ( Settings:           TCanIfSettings );                 (* Device/proto settings *)
         );
  end;

  // helpers for can.bcm

  TCanBcmTimeVal = record      // high level for bcm_timeval in can.bcm
    Sec:  clong;
    uSec: clong;
  end;

  {$PACKENUM 4}                 // we need this type to be 32 bits
  TCanBcmOpCode = (             // high level for bcm message head opcodes
    bopTxSetup   = TX_SETUP,    (* create (cyclic) transmission task                 *)
    bopTxDelete  = TX_DELETE,   (* remove (cyclic) transmission task                 *)
    bopTxRead    = TX_READ,     (* read properties of (cyclic) transmission task     *)
    bopTxSend    = TX_SEND,     (* send one CAN frame                                *)
    bopRxSetup   = RX_SETUP,    (* create RX content filter subscription             *)
    bopRxDelete  = RX_DELETE,   (* remove RX content filter subscription             *)
    bopRxRead    = RX_READ,     (* read properties of RX content filter subscription *)
    bopTxStatus  = TX_STATUS,   (* reply to TX_READ request                          *)
    bopTxExpired = TX_EXPIRED,  (* notification on performed transmissions (count=0) *)
    bopRxStatus  = RX_STATUS,   (* reply to RX_READ request                          *)
    bopRxTimeout = RX_TIMEOUT,  (* cyclic message is absent                          *)
    bopRxChanged = RX_CHANGED); (* updated CAN frame (detected content change)       *)
  {$PACKENUM DEFAULT}

  TCanBcmFlag = (
    bfSetTimer,                 // SETTIMER           = $0001;
    bfStartTimer,               // STARTTIMER         = $0002;
    bfTxCountEVT,               // TX_COUNTEVT        = $0004;
    bfTxAnnounce,               // TX_ANNOUNCE        = $0008;
    bfTxCpCanID,                // TX_CP_CAN_ID       = $0010;
    bfRxFilterID,               // RX_FILTER_ID       = $0020;
    bfRxCheckDLC,               // RX_CHECK_DLC       = $0040;
    bfRxNoAutoTimer,            // RX_NO_AUTOTIMER    = $0080;
    bfRxAnnounceResume,         // RX_ANNOUNCE_RESUME = $0100;
    bfTxResetMultiIDX,          // TX_RESET_MULTI_IDX = $0200;
    bfRxRtrFrame,               // RX_RTR_FRAME       = $0400;
    bfCanFDFrame);              // CAN_FD_FRAME       = $0800;

  {$PACKSET 4}                  // we need this set to be stored in 32 bits like in C
  TCanBcmFlagSet = set of TCanBcmFlag;
  {$PACKSET DEFAULT}

  TCanBcmMsgHead = record             // high level for bcm_msg_head_without_can_frame in can.bcm
    case byte of                      // use this record when you need more then 1 CAN frame in BCM
      1: ( __OpCode:  cuint32;        // keep C like access (__OpCode := TX_SETUP)
           __Flags:   cuint32; );     // keep C like access (__Flags  := SETTIMER OR STARTTIMER OR TX_ANNOUNCE)
      2: (   OpCode:  TCanBcmOpCode;  // Pascalized access  (  OpCode := bopTxSetup)
             Flags:   TCanBcmFlagSet; // Pascalized access  (  Flags  := [bfSetTimer, bfStartTimer, bfTxAnnounce])
             Count:   cuint32;
             Time1:   TCanBcmTimeVal; // was iVal1 in C structure
             Time2:   TCanBcmTimeVal; // was iVal2 in C structure
             ID:      TCanID;
             NFrames: cuint32;)
             // Frame:  array[0..0] of TCanFrame; // Frames[] removed because for BCM cyclic demo when more then 1 frame is needed we already
  end;                                             // had Frames[0] in original bcm_msg_head_can_frame therefore we had to create 2 frame structures

  TCanBcmMsg = record            // high level for bcm_msg_head_can_frame in can.bcm
    Head:  TCanBcmMsgHead;
    Frame: array[0..0] of TCanFrame;
  end;

  TCanFDBcmMsg = record          // high level for bcm_msg_head_canfd_frame in can.bcm
    Head:  TCanBcmMsgHead;
    Frame: array[0..0] of TCanFDFrame;
  end;

  TCanBcmMsg10Frames = record    // holds up to 10 CAN frames, helps to avoid forcing user to create his own message record holding TCanBcmMsgHead and array[0..LessThen10] of TCanFrame
    Head:  TCanBcmMsgHead;
    Frame: array[0..9] of TCanFrame;
  end;

  TCanFDBcmMsg10Frames = record  // holds up to 10 CAN FD frames, helps to avoid forcing user to create his own message record holding TCanFDBcmMsgHead and array[0..LessThen10] of TCanFDFrame
    Head:  TCanBcmMsgHead;
    Frame: array[0..9] of TCanFDFrame;
  end;

  // Protocol families helper, as defined in socketsh.inc
  TProtocolFamily = ( pfUnspec     = PF_UNSPEC,
                      pfUnix       = PF_UNIX,
                      pfLocal      = PF_LOCAL,
                      pfInet       = PF_INET,
                      pfAX25       = PF_AX25,
                      pfIPX        = PF_IPX,
                      pfAppleTalk  = PF_APPLETALK,
                      pfNetRom     = PF_NETROM,
                      pfBridge     = PF_BRIDGE,
                      pfATMPVC     = PF_ATMPVC,
                      pfX25        = PF_X25,
                      pfInet6      = PF_INET6,
                      pfRose       = PF_ROSE,
                      pfDecNet     = PF_DECnet,
                      pfNetBEUI    = PF_NETBEUI,
                      pfSecurity   = PF_SECURITY,
                      pfKey        = PF_KEY,
                      pfNetLink    = PF_NETLINK,
                      pfRoute      = PF_ROUTE,
                      pfPacket     = PF_PACKET,
                      pfASH        = PF_ASH,
                      pfEcoNet     = PF_ECONET,
                      pfATMSVC     = PF_ATMSVC,
                      pfRDS        = PF_RDS,
                      pfSNA        = PF_SNA,
                      pfIRDA       = PF_IRDA,
                      pfPPOX       = PF_PPPOX,
                      pfWanPipe    = PF_WANPIPE,
                      pfLLC        = PF_LLC,
                      pfIB         = PF_IB,
                      pfMPLS       = PF_MPLS,
                      pfCAN        = PF_CAN,
                      pfTIPC       = PF_TIPC,
                      pfBluetooth  = PF_BLUETOOTH,
                      pfIUCV       = PF_IUCV,
                      pfRXRPC      = PF_RXRPC,
                      pfISDN       = PF_ISDN,
                      pfPHONET     = PF_PHONET,
                      pfIEEE802154 = PF_IEEE802154,
                      pfCAIF       = PF_CAIF,
                      pfALG        = PF_ALG,
                      pfNFC        = PF_NFC,
                      pfVSOCK      = PF_VSOCK,
                      pfKCM        = PF_KCM,
                      pfQIPCRTR    = PF_QIPCRTR,
                      pfSMC        = PF_SMC,
                      pfXDP        = PF_XDP,
                      pfMAX        = PF_MAX );

  TCanHandle = cint;

  // helpers for socketsh.inc
  function  CanSocket      (domain: cint; xtype: cint; protocol: cint): TCanHandle; overload; inline;
  function  CanSocket      (domain: TProtocolFamily; xtype: cint; protocol: cint): TCanHandle; overload; inline;
  function  CanSocket      (xtype: cint; protocol: cint): TCanHandle; overload; inline;
  function  CanRecv        (Handle: TCanHandle; buf: pointer; len: size_t; flags: cint): ssize_t; inline;
  function  CanRecvFrom    (Handle: TCanHandle; buf: pointer; len: size_t; flags: cint; from: psockaddr; fromlen : psocklen): ssize_t;
  function  CanSendTo      (Handle: TCanHandle; msg: pointer; len: size_t; flags: cint; tox: psockaddr; tolen: tsocklen): ssize_t; inline;
  function  CanBind        (Handle: TCanHandle; addrx: psockaddr; addrlen: tsocklen): cint; overload; inline;
  function  CanBind        (Handle: TCanHandle; CanSockAddr: TCanSockAddress; addrlen: tsocklen): cint; overload; inline;
  function  CanBind        (Handle: TCanHandle; CanSockAddr: TCanSockAddress): cint; overload; inline;
  function  CanListen      (Handle: TCanHandle; backlog: cint): cint; inline;
  function  CanAccept      (Handle: TCanHandle; addrx: psockaddr; addrlen: psocklen): cint; inline;
  function  CanConnect     (Handle: TCanHandle; name: psockaddr; namelen: tsocklen): cint; inline; overload;
  function  CanConnect     (Handle: TCanHandle; CanSockAddr: TCanSockAddress; namelen: tsocklen): cint; inline; overload;
  function  CanConnect     (Handle: TCanHandle; CanSockAddr: TCanSockAddress): cint; inline; overload;
  function  CanShutdown    (Handle: TCanHandle; how: cint): cint; inline;
  function  CanGetSockName (Handle: TCanHandle; name: psockaddr; namelen: psocklen): cint; inline;
  function  CanGetPeerName (Handle: TCanHandle; name: psockaddr; namelen: psocklen): cint; inline;
  function  CanGetSockOpt  (Handle: TCanHandle; level: cint; optname: cint; optval: pointer; optlen: psocklen): cint; inline; overload;
  function  CanGetSockOpt  (Handle: TCanHandle; level: cint; optname: cint; var optval; optlen: psocklen): cint; inline; overload;
  function  CanSetSockOpt  (Handle: TCanHandle; level: cint; optname: cint; optval: pointer; optlen: tsocklen): cint; inline; overload;
  function  CanSetSockOpt  (Handle: TCanHandle; level: cint; optname: cint; var optval; optlen: tsocklen): cint; inline; overload;
  function  CanSocketPair  (d: cint; xtype: cint; protocol: cint; sv: pcint): cint; inline;

  // helpers for bunxh.inc
  function  CanIOCtl       (Handle: TCanHandle; Ndx: TIOCtlRequest; Data: Pointer): cint; inline; overload;
  function  CanIOCtl       (Handle: TCanHandle; Ndx: TIOCtlRequest; var Data): cint; inline; overload;
  function  CanClose       (Handle: TCanHandle): cInt; inline;

  // helpers for bunxovlh.inc
  function  CanRead        (Handle: TCanHandle; var buf; nbytes : TSize): TSsize; inline;
  function  CanRead        (Handle: TCanHandle; var CanFrame: TCanFrame): TSsize; inline;
  function  CanRead        (Handle: TCanHandle; var CanFDFrame: TCanFDFrame): TSsize; inline;
  function  CanWrite       (Handle: TCanHandle; const buf; nbytes : TSize): TSsize; inline; overload;
  function  CanWrite       (Handle: TCanHandle; const CanFrame: TCanFrame): TSsize; inline; overload;
  function  CanWrite       (Handle: TCanHandle; const CanFDFrame: TCanFDFrame): TSsize; inline; overload;
  function  CanFcntl       (Handle: TCanHandle; Cmd: cint): cint; inline; overload;
  function  CanFcntl       (Handle: TCanHandle; Cmd, Arg: cint): cint; inline; overload;
  function  CanGetErrNo    : longint; inline;

implementation

procedure TCanFrame.Clear;
begin
  Self := Default(TCanFrame);
end;

class operator TCanFrame.Initialize(var ACanFrame: TCanFrame);
begin
  ACanFrame := Default(TCanFrame);
end;

procedure TCanFDFrame.Clear;
begin
  Self := Default(TCanFDFrame);
end;

class operator TCanFDFrame.Initialize(var ACanFDFrame: TCanFDFrame);
begin
  ACanFDFrame := Default(TCanFDFrame);
end;

procedure TCanSockAddress.Clear;
begin
  Self := Default(TCanSockAddress);
end;

class operator TCanSockAddress.Initialize(var ACanSockAddress: TCanSockAddress);
begin
  ACanSockAddress := Default(TCanSockAddress);
end;

function TCanIFreq.GetName: string;
begin
  Result := StrPas(@Self.CName[0]);
end;

procedure TCanIFreq.SetName(const AName: string);
begin
  StrPCopy(@Self.CName[0], LeftStr(AName, IFNAMSIZ - 1));
end;

function  TCanErrMask.GetSFF: boolean;
begin
  Result := Self.Value = CAN_SFF_MASK;
end;

procedure TCanErrMask.SetSFF(const ANewMaskSFF: boolean);
begin
  if ANewMaskSFF then
    Self.Value := CAN_SFF_MASK
  else
    Self.Value := 0;
end;

function  TCanErrMask.GetEFF: boolean;
begin
  Result := Self.Value = CAN_EFF_MASK;
end;

procedure TCanErrMask.SetEFF(const ANewMaskEFF: boolean);
begin
  if ANewMaskEFF then
    Self.Value := CAN_EFF_MASK
  else
    Self.Value := 0;
end;

function CanSocket(domain: cint; xtype: cint; protocol: cint): TCanHandle;
begin
  Result := fpsocket(domain, xtype, protocol);
end;

function CanSocket(domain: TProtocolFamily; xtype: cint; protocol: cint): TCanHandle;
begin
  Result := fpsocket(Ord(domain), xtype, protocol);
end;

function CanSocket(xtype: cint; protocol: cint): TCanHandle;
begin
  Result := fpsocket(PF_CAN, xtype, protocol);
end;

function CanRecv(Handle: TCanHandle; buf: pointer; len: size_t; flags: cint): ssize_t;
begin
  Result := fprecv(Handle, buf, len, flags);
end;

function CanRecvFrom(Handle: TCanHandle; buf: pointer; len: size_t; flags: cint; from: psockaddr; fromlen : psocklen): ssize_t;
begin
  Result := fprecvfrom(Handle, buf, len, flags, from, fromlen);
end;

function CanSend(Handle: TCanHandle; msg: pointer; len: size_t; flags: cint): ssize_t;
begin
  Result := fpsend(Handle, msg, len, flags);
end;

function CanSendTo(Handle: TCanHandle; msg: pointer; len: size_t; flags: cint; tox: psockaddr; tolen: tsocklen): ssize_t;
begin
  Result := fpsendto(Handle, msg, len, flags, tox, tolen);
end;

function CanBind(Handle: TCanHandle; addrx: psockaddr; addrlen: tsocklen): cint;
begin
  Result := fpbind(Handle, addrx, addrlen);
end;

function CanBind(Handle: TCanHandle; CanSockAddr: TCanSockAddress; addrlen: tsocklen): cint;
begin
  Result := fpbind(Handle, @CanSockAddr, addrlen);
end;

function CanBind(Handle: TCanHandle; CanSockAddr: TCanSockAddress): cint;
begin
  Result := fpbind(Handle, @CanSockAddr, SizeOf(CanSockAddr));
end;

function CanListen(Handle: TCanHandle; backlog: cint): cint;
begin
  Result := fplisten(Handle, backlog);
end;

function CanAccept(Handle: TCanHandle; addrx: psockaddr; addrlen: psocklen): cint;
begin
  Result := fpaccept(Handle, addrx, addrlen);
end;

function CanConnect(Handle: TCanHandle; name: psockaddr; namelen: tsocklen): cint;
begin
  Result := fpconnect(Handle, name, namelen);
end;

function CanConnect(Handle: TCanHandle; CanSockAddr: TCanSockAddress; namelen: tsocklen): cint; inline; overload;
begin
  Result := fpconnect(Handle, @CanSockAddr, namelen);
end;

function CanConnect(Handle: TCanHandle; CanSockAddr: TCanSockAddress): cint; inline; overload;
begin
  Result := fpconnect(Handle, @CanSockAddr, SizeOf(CanSockAddr));
end;

function CanShutdown(Handle: TCanHandle; how: cint): cint;
begin
  Result := fpshutdown(Handle, how);
end;

function CanGetSockName(Handle: TCanHandle; name: psockaddr; namelen: psocklen): cint;
begin
  Result := fpgetsockname(Handle, name, namelen);
end;

function CanGetPeerName(Handle: TCanHandle; name: psockaddr; namelen: psocklen): cint;
begin
  Result := fpgetpeername(Handle, name, namelen);
end;

function CanGetSockOpt(Handle: TCanHandle; level: cint; optname: cint; optval: pointer; optlen: psocklen): cint;
begin
  Result := fpgetsockopt(Handle, level, optname, optval, optlen);
end;

function CanGetSockOpt(Handle: TCanHandle; level: cint; optname: cint; var optval; optlen: psocklen): cint;
begin
  Result := fpgetsockopt(Handle, level, optname, @optval, optlen);
end;

function CanSetSockOpt(Handle: TCanHandle; level: cint; optname: cint; optval: pointer; optlen: tsocklen): cint;
begin
  Result := fpsetsockopt(Handle, level, optname, optval, optlen);
end;

function CanSetSockOpt(Handle: TCanHandle; level: cint; optname: cint; var optval; optlen: tsocklen): cint;
begin
  Result := fpsetsockopt(Handle, level, optname, @optval, optlen);
end;

function CanSocketPair(d: cint; xtype: cint; protocol: cint; sv: pcint): cint;
begin
  Result := fpsocketpair(d, xtype, protocol, sv);
end;

function CanIOCtl(Handle: TCanHandle; Ndx: TIOCtlRequest; Data: Pointer): cint;
begin
  Result := fpioctl(Handle, Ndx, Data);
end;

function CanIOCtl(Handle: TCanHandle; Ndx: TIOCtlRequest; var Data): cint;
begin
  Result := fpioctl(Handle, Ndx, @Data);
end;

function CanClose(Handle : TCanHandle): cInt;
begin
  Result := fpclose(Handle);
end;

function CanRead(Handle: TCanHandle; var buf; nbytes : TSize): TSsize;
begin
  Result := fpread(Handle, buf, nbytes);
end;

function CanRead(Handle: TCanHandle; var CanFrame: TCanFrame): TSsize;
begin
  Result := fpread(Handle, @CanFrame, SizeOf(CanFrame));
end;

function CanRead(Handle: TCanHandle; var CanFDFrame: TCanFDFrame): TSsize;
begin
  Result := fpread(Handle, @CanFDFrame, SizeOf(CanFDFrame));
end;

function CanWrite(Handle: TCanHandle; const buf; nbytes : TSize): TSsize;
begin
  Result := fpwrite(Handle, buf, nbytes);
end;

function CanWrite(Handle: TCanHandle; const CanFrame: TCanFrame): TSsize;
begin
  Result := fpwrite(Handle, @CanFrame, SizeOf(CanFrame));
end;

function CanWrite(Handle: TCanHandle; const CanFDFrame: TCanFDFrame): TSsize;
begin
  Result := fpwrite(Handle, @CanFDFrame, SizeOf(CanFDFrame));
end;

function CanFcntl(Handle: TCanHandle; Cmd: cint): cint;
begin
  Result := FpFcntl(Handle, Cmd);
end;

function CanFcntl(Handle: TCanHandle; Cmd, Arg: cint): cint;
begin
  Result := FpFcntl(Handle, Cmd, Arg);
end;

function CanGetErrNo: longint;
begin
  Result := fpgeterrno;
end;

end.
