////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                //
//  hlcanerrsim utility to simulate SocketCAN error messages, by Zeljko Avramovic (c) 2021        //
//                                                                                                //
//  Dual license:                                                                                 //
//  1. FPC modified LGPL (chosen for compatibility with FreePascal and Lazarus)                   //
//  2. BSD3              (chosen for compatibility with everything else)                          //
//                                                                                                //
//  Virtual CAN adapter vcan0 is hard coded and you can bring it up like this:                    //
//  sudo modprobe vcan                                                                            //
//  sudo ip link add dev vcan0 type vcan                                                          //
//  sudo ip link set vcan0 mtu 72              # needed for CAN FD                                //
//  sudo ip link set vcan0 up                                                                     //
//                                                                                                //
//  To simulate error messages use hlcanerrsim utility like this:                                 //
//  ./hlcanerrsim LostArbitrationBit=09 TX BusOff Data7=AA Data6=BB                               //
//                                                                                                //
//  That should show in hlcanerrdump utility as:                                                  //
//  0x04A [8] 16 00 80 00 00 00 BB AA  ERR=LostArbitrationBit09,BusOff,Prot(Type(TX),Loc(Unspec)) //
//                                                                                                //
//  Alternatively, you could use candump from can-utils to check error messages like this:        //
//  candump -s -c -c -x -a any,0~0,#FFFFFFFF                                                      //
//                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////

program hlcanerrsim; // CAN error messages simulator

{$mode objfpc} {$H+}

uses
  sysutils, strutils, sockets, can, can.hl, can.error;

procedure ShowHelpAndExit;
begin
  WriteLn('');
  WriteLn('Usage: hlcanerrsim <CAN interface> [options]');
  WriteLn('');
  WriteLn('CAN interface:                   ( CAN interface is case sensitive )');
  WriteLn('    can1                         ( or can2, can3 or virtual ones like vcan0, vcan1...');
  WriteLn('');
  WriteLn('Options:                         ( options are not case sensitive )');
  WriteLn('                                 ( ERROR CLASS (MASK) IN CAN ID: )');
  WriteLn('    TxTimeout                    ( TX timeout by netdevice driver )');
  WriteLn('    NoAck                        ( received no ACK on transmission )');
  WriteLn('    BusOff                       ( bus off )');
  WriteLn('    BusError                     ( bus error, may flood! )');
  WriteLn('    Restarted                    ( controller restarted )');
  WriteLn('                                 ( ARBITRATIONLOST IN CAN ID + BIT NUMBER IN DATA[0]: )');
  WriteLn('    LostArbitrationBit=<00..29>  ( decimal bit number in bitstream )');
  WriteLn('                                 ( CONTROLLER IN CAN ID + ERROR STATUS IN DATA[1]: )');
  WriteLn('    OverflowRX                   ( RX buffer overflow )');
  WriteLn('    OverflowTX                   ( TX buffer overflow )');
  WriteLn('    WarningRX                    ( reached warning level for RX errors )');
  WriteLn('    WarningTX                    ( reached warning level for TX errors )');
  WriteLn('    PassiveRX                    ( reached error passive status RX, errors > 127 )');
  WriteLn('    PassiveTX                    ( reached error passive status TX, errors > 127 )');
  WriteLn('    Active                       ( recovered to error active state )');
  WriteLn('                                 ( PROTOCOL ERROR IN CAN ID + TYPE IN DATA[2]: )');
  WriteLn('    SingleBit                    ( single bit error )');
  WriteLn('    FrameFormat                  ( frame format error )');
  WriteLn('    BitStuffing                  ( bit stuffing error )');
  WriteLn('    Bit0                         ( unable to send dominant bit )');
  WriteLn('    Bit1                         ( unable to send recessive bit )');
  WriteLn('    BusOverload                  ( bus overload )');
  WriteLn('    ActiveAnnouncement           ( active error announcement )');
  WriteLn('    TX                           ( error occurred on transmission )');
  WriteLn('                                 ( PROTOCOL ERROR IN CAN ID + LOCATION IN DATA[3]: )');
  WriteLn('    SOF                          ( start of frame )');
  WriteLn('    ID28_21                      ( ID bits 21..28, SFF: 3..10 )');
  WriteLn('    ID20_18                      ( ID bits 18..20, SFF: 0..2 )');
  WriteLn('    SRTR                         ( substitute RTR, SFF: RTR )');
  WriteLn('    IDE                          ( identifier extension )');
  WriteLn('    ID17_13                      ( ID bits 13..17 )');
  WriteLn('    ID12_05                      ( ID bits 5..12 )');
  WriteLn('    ID04_00                      ( ID bits 0..4 )');
  WriteLn('    RTR                          ( RTR )');
  WriteLn('    RES1                         ( reserved bit 1 )');
  WriteLn('    RES0                         ( reserved bit 0 )');
  WriteLn('    DLC                          ( data length code )');
  WriteLn('    DATA                         ( data section )');
  WriteLn('    CRC_SEQ                      ( CRC sequence )');
  WriteLn('    CRC_DEL                      ( CRC delimiter )');
  WriteLn('    ACK                          ( ACK slot )');
  WriteLn('    ACK_DEL                      ( ACK delimiter )');
  WriteLn('    EOF                          ( end of frame )');
  WriteLn('    INTERM                       ( intermission )');
  WriteLn('                                 ( TRANSCEIVER ERROR IN CAN ID + STATUS IN DATA[4]: )');
  WriteLn('                                 ( CANH CANL )');
  WriteLn('    TransUnspec                  ( 0000 0000 )');
  WriteLn('    CanHiNoWire                  ( 0000 0100 )');
  WriteLn('    CanHiShortToBAT              ( 0000 0101 )');
  WriteLn('    CanHiShortToVCC              ( 0000 0110 )');
  WriteLn('    CanHiShortToGND              ( 0000 0111 )');
  WriteLn('    CanLoNoWire                  ( 0100 0000 )');
  WriteLn('    CanLoShortToBAT              ( 0101 0000 )');
  WriteLn('    CanLoShortToVCC              ( 0110 0000 )');
  WriteLn('    CanLoShortToGND              ( 0111 0000 )');
  WriteLn('    CanLoShortToCanHi            ( 1000 0000 )');
  WriteLn('                                 ( CUSTOM BYTE TO DATA[0..7]: )');
  WriteLn('    Data<0..7>=<00..FF>          ( write hex number to one of 8 payload bytes )');
  WriteLn('');
  WriteLn('Examples:');
  WriteLn('');
  WriteLn('    ./hlcanerrsim can1 LostArbitrationBit=09 TX BusOff Data7=AA Data6=BB');
  WriteLn('    ( can1: 9th arb. bit lost, transmission, bus off, controller specific info in Data[6] and Data[7] )');
  WriteLn('');
  WriteLn('    ./hlcanerrsim vcan0 NoAck TxTimeout Active');
  WriteLn('    ( vcan0: received no ACK on transmission, driver timeout, protocol type active error announcement )');
  WriteLn('');
  WriteLn('    ./hlcanerrsim vcan0 BusError CanHiNoWire Restarted INTERM');
  WriteLn('    ( vcan0: bus error, lost CANH wiring, controller restarted, protocol location intermission )');
  WriteLn('');
  Halt;
end;

procedure ShowErrAndExit(ErrType: string);
begin
  WriteLn('Error: You can only have one ' + ErrType + ' parameter!');
  Halt;
end;

procedure ShowLocErrAndExit;
begin
  ShowErrAndExit('protocol location');
end;

procedure ShowArbErrAndExit;
begin
  ShowErrAndExit('arbitration bit');
end;

procedure ShowTranscErrAndExit;
begin
  ShowErrAndExit('transceiver');
end;

procedure perror(const S : string);
begin
  WriteLn(S, ', SocketError = ', SocketError);
end;

var
  i:                    integer;
  CanHandle:            TCanHandle;
  CanAddress:           TCanSockAddress;
  CanInterface:         TCanIfreq;
  CanFrame:             TCanFrame;
  ArbitrationProcessed: boolean = false;
  LocationProcessed:    boolean = false;
  TransceiverProcessed: boolean = false;
  ParStr:               string;
begin
  try
    WriteLn('CAN Sockets Error Messages Simulator');
    if ParamCount() = 0 then
      ShowHelpAndExit
    else
    begin
    	// WriteLn(ParamStr(0));
      CanFrame.Clear;
      CanFrame.ID.ERR := true;
      CanFrame.DLC    := CAN_ERR_DLC; // error frame is always 8 bytes
      for i := 2 to ParamCount() do
  	  begin // parse command line parameters
        ParStr := UpperCase(ParamStr(i));
        case ParStr of
          'TXTIMEOUT':          CanFrame.ID.Mask.Error.TxTimeout                          := true;
          'NOACK':              CanFrame.ID.Mask.Error.NoAck                              := true;
          'BUSOFF':             CanFrame.ID.Mask.Error.BusOff                             := true;
          'BUSERROR':           CanFrame.ID.Mask.Error.BusError                           := true;
          'RESTARTED':          CanFrame.ID.Mask.Error.Restarted                          := true;
          'OVERFLOWRX':         begin
                                  CanFrame.ID.Mask.Error.Controller                       := true;
                                  CanFrame.DataError.Controller.OverflowRX                := true;
                                end;
          'OVERFLOWTX':         begin
                                  CanFrame.ID.Mask.Error.Controller                       := true;
                                  CanFrame.DataError.Controller.OverflowTX                := true;
                                end;
          'WARNINGRX':          begin
                                  CanFrame.ID.Mask.Error.Controller                       := true;
                                  CanFrame.DataError.Controller.WarningRX                 := true;
                                end;
          'WARNINGTX':          begin
                                  CanFrame.ID.Mask.Error.Controller                      := true;
                                  CanFrame.DataError.Controller.WarningTX                 := true;
                                end;
          'PASSIVERX':          begin
                                  CanFrame.ID.Mask.Error.Controller                       := true;
                                  CanFrame.DataError.Controller.PassiveRX                 := true;
                                end;
          'PASSIVETX':          begin
                                  CanFrame.ID.Mask.Error.Controller                       := true;
                                  CanFrame.DataError.Controller.PassiveTX                 := true;
                                end;
          'ACTIVE':             begin
                                  CanFrame.ID.Mask.Error.Controller                       := true;
                                  CanFrame.DataError.Controller.Active                    := true;
                                end;
          'SINGLEBIT':          begin
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Category.SingleBit          := true;
                                end;
          'FRAMEFORMAT':        begin
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Category.FrameFormat        := true;
                                end;
          'BITSTUFFING':        begin
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Category.BitStuffing        := true;
                                end;
          'BIT0':               begin
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Category.Bit0               := true;
                                end;
          'BIT1':               begin
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Category.Bit1               := true;
                                end;
          'BUSOVERLOAD':        begin
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Category.BusOverload        := true;
                                end;
          'ACTIVEANNOUNCEMENT': begin
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Category.ActiveAnnouncement := true;
                                end;
          'TX':                 begin
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Category.TX                 := true;
                                end;
          'LOCUNSPEC':          begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_UNSPEC;
                                  LocationProcessed                                       := true;
                                end;
          'SOF':                begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_SOF;
                                  LocationProcessed                                       := true;
                                end;
          'ID28_21':            begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_ID28_21;
                                  LocationProcessed                                       := true;
                                end;
          'ID20_18':            begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_ID20_18;
                                  LocationProcessed                                       := true;
                                end;
          'SRTR':               begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_SRTR;
                                  LocationProcessed                                       := true;
                                end;
          'IDE':                begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_IDE;
                                  LocationProcessed                                       := true;
                                end;
          'ID17_13':            begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_ID17_13;
                                  LocationProcessed                                       := true;
                                end;
          'ID12_05':            begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_ID12_05;
                                  LocationProcessed                                       := true;
                                end;
          'ID04_00':            begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_ID04_00;
                                  LocationProcessed                                       := true;
                                end;
          'RTR':                begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_RTR;
                                  LocationProcessed                                       := true;
                                end;
          'RES1':               begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_RES1;
                                  LocationProcessed                                       := true;
                                end;
          'RES0':               begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_RES0;
                                  LocationProcessed                                       := true;
                                end;
          'DLC':                begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_DLC;
                                  LocationProcessed                                       := true;
                                end;
          'DATA':               begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_DATA;
                                  LocationProcessed                                       := true;
                                end;
          'CRC_SEQ':            begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_CRC_SEQ;
                                  LocationProcessed                                       := true;
                                end;
          'CRC_DEL':            begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_CRC_DEL;
                                  LocationProcessed                                       := true;
                                end;
          'ACK':                begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_ACK;
                                  LocationProcessed                                       := true;
                                end;
          'ACK_DEL':            begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_ACK_DEL;
                                  LocationProcessed                                       := true;
                                end;
          'EOF':                begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_EOF;
                                  LocationProcessed                                       := true;
                                end;
          'INTERM':             begin
                                  if LocationProcessed then ShowLocErrAndExit;
                                  CanFrame.ID.Mask.Error.Protocol                         := true;
                                  CanFrame.DataError.Protocol.Location                    := CAN_ERR_PROT_LOC_INTERM;
                                  LocationProcessed                                       := true;
                                end;
          'TRANSUNSPEC':        begin
                                  if TransceiverProcessed then ShowTranscErrAndExit;
                                  CanFrame.ID.Mask.Error.Transceiver                      := true;
                                  CanFrame.DataError.Transceiver                          := CAN_ERR_TRX_UNSPEC;
                                  TransceiverProcessed                                    := true;
                                end;
          'CANHINOWIRE':        begin
                                  if TransceiverProcessed then ShowTranscErrAndExit;
                                  CanFrame.ID.Mask.Error.Transceiver                      := true;
                                  CanFrame.DataError.Transceiver                          := CAN_ERR_TRX_CANH_NO_WIRE;
                                  TransceiverProcessed                                    := true;
                                end;
          'CANHISHORTTOBAT':    begin
                                  if TransceiverProcessed then ShowTranscErrAndExit;
                                  CanFrame.ID.Mask.Error.Transceiver                      := true;
                                  CanFrame.DataError.Transceiver                          := CAN_ERR_TRX_CANH_SHORT_TO_BAT;
                                  TransceiverProcessed                                    := true;
                                end;
          'CANHISHORTTOVCC':    begin
                                  if TransceiverProcessed then ShowTranscErrAndExit;
                                  CanFrame.ID.Mask.Error.Transceiver                      := true;
                                  CanFrame.DataError.Transceiver                          := CAN_ERR_TRX_CANH_SHORT_TO_VCC;
                                  TransceiverProcessed                                    := true;
                                end;
          'CANHISHORTTOGND':    begin
                                  if TransceiverProcessed then ShowTranscErrAndExit;
                                  CanFrame.ID.Mask.Error.Transceiver                      := true;
                                  CanFrame.DataError.Transceiver                          := CAN_ERR_TRX_CANH_SHORT_TO_GND;
                                  TransceiverProcessed                                    := true;
                                end;
          'CANLONOWIRE':        begin
                                  if TransceiverProcessed then ShowTranscErrAndExit;
                                  CanFrame.ID.Mask.Error.Transceiver                      := true;
                                  CanFrame.DataError.Transceiver                          := CAN_ERR_TRX_CANL_NO_WIRE;
                                  TransceiverProcessed                                    := true;
                                end;
          'CANLOSHORTTOBAT':    begin
                                  if TransceiverProcessed then ShowTranscErrAndExit;
                                  CanFrame.ID.Mask.Error.Transceiver                      := true;
                                  CanFrame.DataError.Transceiver                          := CAN_ERR_TRX_CANL_SHORT_TO_BAT;
                                  TransceiverProcessed                                    := true;
                                end;
          'CANLOSHORTTOVCC':    begin
                                  if TransceiverProcessed then ShowTranscErrAndExit;
                                  CanFrame.ID.Mask.Error.Transceiver                      := true;
                                  CanFrame.DataError.Transceiver                          := CAN_ERR_TRX_CANL_SHORT_TO_VCC;
                                  TransceiverProcessed                                    := true;
                                end;
          'CANLOSHORTTOGND':    begin
                                  if TransceiverProcessed then ShowTranscErrAndExit;
                                  CanFrame.ID.Mask.Error.Transceiver                      := true;
                                  CanFrame.DataError.Transceiver                          := CAN_ERR_TRX_CANL_SHORT_TO_GND;
                                  TransceiverProcessed                                    := true;
                                end;
          'CANLOSHORTTOCANHI':  begin
                                  if TransceiverProcessed then ShowTranscErrAndExit;
                                  CanFrame.ID.Mask.Error.Transceiver                      := true;
                                  CanFrame.DataError.Transceiver                          := CAN_ERR_TRX_CANL_SHORT_TO_CANH;
                                  TransceiverProcessed                                    := true;
                                end;
        otherwise // case else
          if (ParStr.Length = 'LostArbitrationBit=00'.Length)             and
             (LeftStr(ParStr, ParStr.Length - 2) = 'LOSTARBITRATIONBIT=') and
             (ParStr[ParStr.Length - 1]     in ['0'..'2'])                and // valid bits are from 00 to 29 (in decimal)
             (ParStr[ParStr.Length]         in ['0'..'9'])               then // valid bits are from 00 to 29 (in decimal)
          begin
            if ArbitrationProcessed then ShowArbErrAndExit;
            CanFrame.ID.Mask.Error.LostArbitration       := true;
            CanFrame.DataError.LostArbitration.BitNumber := StrToInt(RightStr(ParStr, 2));
            ArbitrationProcessed := true;
            // WriteLn('LostArbitrationBit=', StrToInt(RightStr(ParStr, 2)));
          end
          else if ((ParStr.Length      =  'Data1=00'.Length) and
                   (LeftStr(ParStr, 4) =  'DATA')             and
                   (ParStr[5]          in ['0'..'7'])         and // valid data bytes are 0 to 7
                   (ParStr[6]          =  '=')                and
                   (ParStr[7]          in ['0'..'F'])         and // valid hex nibbles are from 0 to F
                   (ParStr[8]          in ['0'..'F']))       then // valid hex nibbles are from 0 to F
          begin
            CanFrame.Data[StrToInt(ParStr[5])] := StrToInt('$' + RightStr(ParStr, 2));
            //WriteLn(CanFrame.DataQWord[0].ToHexString);
          end
          else
            ShowHelpAndExit;
          {WriteLn(ParStr.Length, ' ', 'LostArbitrationBit00'.Length, ' ',
                  (LeftStr(ParStr, 18) = 'LOSTARBITRATIONBIT'), ' ',
                  ParStr[ParStr.Length - 1] in ['0'..'2'], ' ',
                  ParStr[ParStr.Length - 1] in ['0'..'9']);}
          {WriteLn(ParStr.Length       =  'Data1=00'.Length, ' ',
                  (LeftStr(ParStr, 4) =  'DATA'), ' ',
                  ParStr[5]           in ['0'..'7'], ' ',
                  ParStr[6]           =  '=', ' ',
                  ParStr[7]           in ['0'..'F'], ' ',
                  ParStr[8]           in ['0'..'F']);}
        end; //case end
        // WriteLn(i:2, '. argument: ', paramStr(i));
  	  end;
    end;

    CanHandle := CanSocket(SOCK_RAW, CAN_RAW);
    if CanHandle < 0 then
    begin
      perror('Socket');
      Exit;
    end;

    CanInterface.Name := ParamStr(1); // 'vcan0';
    if CanIOCtl(CanHandle, SIOCGIFINDEX, CanInterface) < 0 then
    begin
      perror('IOctl unable to open CAN interface ' + CanInterface.Name + ' (did you set CAN interface as first parameter?)');
      Exit;
    end;

    CanAddress.CanFamily  := AF_CAN;
    CanAddress.CanIfIndex := CanInterface.IfIndex;

    if CanBind(CanHandle, CanAddress) < 0 then
    begin
      perror('Bind');
      Exit;
    end;

    if CanFrame.ID.Full = CAN_ERR_FLAG then
    begin
      WriteLn('Error: Setting just ERR bit in CAN ID is not enough to form a proper error message!');
      Exit;
    end;

    if CanWrite(CanHandle, CanFrame) <> SizeOf(CanFrame) then
    begin
      perror('Write');
      Exit;
    end;
    WriteLn('Sent CAN error message to ' + CanInterface.Name + ' with ID=$', CanFrame.ID.Full.ToHexString, ' and DATA=$', CanFrame.DataQWord[0].ToHexString);

    if CanClose(CanHandle) < 0 then
    begin
      perror('Close');
      Exit;
    end;

  except
    on e: Exception do
      WriteLn(e.Message);
  end;
end.

