//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//  This is a SocketCAN demo converted from C to Pascal by Zeljko Avramovic.    //
//  Original example has GNU General Public License v3.0 and can be found at    //
//  https://github.com/craigpeacock/CAN-Examples                                //
//                                                                              //
//  Virtual CAN adapter vcan0 is hard coded and you can bring it up like this:  //
//  sudo modprobe vcan                                                          //
//  sudo ip link add dev vcan0 type vcan                                        //
//  sudo ip link set vcan0 mtu 72            # needed for CAN FD                //
//  sudo ip link set vcan0 up                                                   //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

program hlcanfilter; // higher level pascalized version of canfilter demo

{$mode delphi} {$H+}

uses
  sysutils, strutils, sockets, can, can.hl, can.raw;

procedure perror(const S : string);
begin
  WriteLn(S, ', SocketError = ', SocketError);
end;

var
  i, BytesNo:   integer;
  CanHandle:    TCanHandle;
  CanAddress:   TCanSockAddress;
  CanInterface: TCanIfreq;
  CanFrame:     TCanFrame;
  CanFilter:    array[0..3] of TCanFilter;
begin
  try
    WriteLn('CAN Sockets Filter Demo');
    CanHandle := CanSocket(SOCK_RAW, CAN_RAW);
    if CanHandle < 0 then
    begin
      perror('Socket');
      Exit;
    end;

    CanInterface.Name := 'vcan0';
    if CanIOCtl(CanHandle, SIOCGIFINDEX, CanInterface) < 0 then
    begin
      perror('IOctl');
      Exit;
    end;

    CanAddress.Clear;
    CanAddress.CanFamily  := AF_CAN;
    CanAddress.CanIfIndex := CanInterface.IfIndex;

    if CanBind(CanHandle, CanAddress) < 0 then
    begin
      perror('Bind');
      Exit;
    end;

    CanFilter[0].ID.Value      := $550;      // 1st CAN ID
    CanFilter[0].Mask.Value    := $7F0;      // 1st CAN MASK matches all frames with identifiers from $550 to $55F
    CanFilter[1].ID.Value      := $200;      // 2nd CAN ID
    CanFilter[1].Mask.Value    := $7FF;      // 2nd CAN MASK matches only frames with identifier equal to $200 and nothing else
    // now let's try high level access for optimized filters found in example at https://www.kernel.org/doc/html/v5.10/networking/can.html
    CanFilter[2].ID.Value      := $123;      // 3rd CAN ID
    CanFilter[2].ID.EFF        := false;     // 3rd CAN ID has standard frame
    CanFilter[2].Mask.SFF      := true;      // 3rd CAN MASK, only SFF frames with $123 CAN ID can pass (cansend vcan0 00000123#11223344)
    CanFilter[2].Mask.Flag.EFF := true;      // 3rd CAN MASK, comment out this line if you want both SFF frames with CAN ID $123 and EFF frames with $XXXXX123 to pass (cansend vcan0 123#11223344 and cansend vcan0 00000123#11223344)
    CanFilter[2].Mask.Flag.RTR := true;      // 3rd CAN MASK, comment out this line if you want both SFF frames with CAN ID $123 and EFF frames with $XXXXX123 to pass (cansend vcan0 123#11223344 and cansend vcan0 00000123#11223344)
    CanFilter[3].ID.Value      := $12345678; // 4th CAN ID
    CanFilter[3].ID.EFF        := true;      // 4th CAN ID has extended frame
    CanFilter[3].Mask.EFF      := true;      // 4th CAN MASK, only EFF frames with $12345678 CAN ID can pass (cansend vcan0 12345678#B01DFACE)
    CanFilter[3].Mask.Flag.EFF := true;      // 4th CAN MASK, only EFF frames with $12345678 CAN ID can pass (cansend vcan0 12345678#B01DFACE)
    CanFilter[3].Mask.Flag.RTR := true;      // 4th CAN MASK, only EFF frames with $12345678 CAN ID can pass (cansend vcan0 12345678#B01DFACE)

    CanSetSockOpt(CanHandle, SOL_CAN_RAW, CAN_RAW_FILTER, CanFilter, SizeOf(CanFilter));

    BytesNo := CanRead(CanHandle, CanFrame);
    if BytesNo < 0 then
    begin
      perror('Read: Error = ' + CanGetErrNo.ToString);
      Exit;
    end;
    if BytesNo < SizeOf(TCanFrame) then // paranoid check
    begin
      perror('Read: Incomplete CAN frame');
      Exit;
    end;

    Write(Format(IfThen(CanFrame.ID.EFF, '0x%08.8X [%d] ', '     0x%03.3X [%d] '), [CanFrame.ID.Value, CanFrame.DLC])); // extended or standard frame

    for i := 0 to CanFrame.DLC - 1 do
      Write(Format('%02.2X ', [CanFrame.Data[i]]));
    WriteLn('');

    if CanClose(CanHandle) < 0 then
    begin
      perror('Close');
      Exit;
    end;

  except
    on e: Exception do
      WriteLn(e.Message);
  end;
end.

